This is a work in progress.

Always use bash, never /bin/sh, etc.

References bash(1)
https://mywiki.wooledge.org/EnglishFrontPage

Note: stackoverflow often has answers that are not robust.
For a common example, see https://mywiki.wooledge.org/ParsingLs

Use common sense and be consistent. For existing scripts, use the
existing style. If you want to change it, put it in a git repo, then
make a commit which only changes the style.

If you read something online that says "this way is more portable", that
often means that bash or gnu can do it in a more readable or otherwise
better way and we want that. We won't be porting our scripts to nongnu
oses or other shells.

Scripts must be executable, preferably with no extension, but .sh is
okay. lowercase-dash-separated. Scripts meant to be sourced (libraries)
must have .sh extension. Rational: look at /usr/bin, scripts don't have
.sh, it's extraneous overhead to the person using it.

Scripts should start like this:
```
#!/bin/bash
# <one line to give the program's name and a brief idea of what it does.>
#
# Copyright (C) 2019 Free Software Foundation
# SPDX-License-Identifier: AGPL-3.0-or-later

if [ -z "$BASH_VERSION" ]; then echo "error: shell is not bash" >&2; exit 1; fi
```

Or if we expect it to actually be read by other people, use the full
copyright notice:

```
#!/bin/bash
# <one line to give the program's name and a brief idea of what it does.>
#
# Copyright (C) 2019 Free Software Foundation
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

if [ -z "$BASH_VERSION" ]; then echo "error: shell is not bash" >&2; exit 1; fi
```

The shebang should not have arguments so bash script-name works the
same. Use the set command instead.


Before using any script, it should have no output from

```
  # 2086 = unquoted $var
  # 2046 = unquoted $(cmd)
  # 2068 = Double quote array expansions to avoid re-splitting elements.
shellcheck -x -e 2086,2046,2068 SCRIPT_FILE
```

For scripts which are expected to take untrusted input, or process files
or variables which could have spaces, run shellcheck with no exceptions
to verify quoting.

If there is a need for something shellcheck warns about, use a directive
to override the warning, eg:

```
# shellcheck disable=SC2035
```

If the shellcheck you are using is too old for -x, get a new one using
package pinning or haskell stack.

Add comments describing functions that are not short and
obvious. Document any arguments, global variables used, return value
that is different from the default of the exit code of the last command
run. See the err script in this repo for example.

Comment any tricky or important parts of code.

Use TODO comments
```
# TODO iank: add x functionality
```

Indent 2 spaces.

Try to wrap lines longer than ~80 chars. Use \ followed by newline where appropriate.

Use here documents to print multiple lines instead multiple echo lines.

cat <<END
I have lots
of stuff to say.
END


Put then and do on the same line as the condition:
```
if x; then
  :
fi
for x; do
  :
done
```

To work from the same directory of the script
```
readonly this_file="$(readlink -f -- "$BASH_SOURCE")"; cd ${this_file%/*}
```

Variables:

Environment variables are all caps. Other vars are lowercase. Underscore
separates words. Declare constants and environment variables at the top
of the file when possible.

```
# Constant
readonly conf_dir=/some/path

# Constant and environment
declare -xr conf_dir=/some/path

# A constant set based on condition needs to be set readonly after it's set.
verbose=false
if [[ $1 ]]; then
  verbose=true
fi
readonly verbose
```

Use local to make function local variables enforced. This adds
readability and avoids error from using variable in an unintended
scope. eg: `local y=5`


Never quote literal integers.

Prefer single quoting strings that have no subsitution (things that
start with $).

Non command options or path names can be optionally quoted. eg `x='oo'`




Functions:

Put all functions together near the top of the file after includes,
constants, and set statements.


Always use `return 0`, never `return`, because you can return nonzero
by accident otherwise.


Error handling:

1. Automatically exit on failed command:

For scripts with no functions add:
set -eE -o pipefail
trap 'echo "$0:$LINENO:error: \"$BASH_COMMAND\" returned $?" >&2' ERR

For scripts with functions, copy/paste or source the err script in this
repository, which will exit on error and print stack traces.

```
source path/to/err
tf() {
  result=true
  echo do stuff
  result=false
}
tf
$result || echo handling false result

echo this prints too
```

To allow a specific command to to fail, make it part of a conditional. eg:
iptables -D rule || [[ $? == 1 ]] # allow exit code 1 when rule doesn't exist

Don't use the return code from your own functions unless its very short
and you are very confident that only one line that can return nonzero,
because automatic error handling ignores any other nonzero errors inside
the function. Instead, use a result var.


2. Less prefered: Manual error checking, but still automatically print failed commands.
Explicit handling should be done for almost all commands not builtin to bash.

```
set -E -o pipefail
trap 'echo "$0:$LINENO:error: \"$BASH_COMMAND\" returned $?" >&2' ERR
```

Or for scripts with functions, as above,

```
source err
err-print
```

Manual error handling example:

```
iptables -D $rule
if [[ $? != [01] ]]; then
  err-exit exiting due to failed iptables
fi
```

3. If doing a best effort and not exiting on errors, exit code should be
the higest error.

```
error=0; trap 'error=$(($?>$error?$?:$error))' ERR
# intentionally ignoring errors to do best effort
command1
command2
command3
exit $error
```

Additional error handling notes:

In some circumstances pipefail may be too blunt. In that case, check
`${PIPESTATUS[@]}`.


Misc:

prefer the use of builtins such as the Parameter Expansion functions in bash(1) as it's more robust.

```
# Prefer this:
addition=$((x + y))
substitution="${string/#foo/bar}"

# Instead of this:
addition="$(expr $x + $y)"
substitution="$(echo "${string}" | sed -e 's/^foo/bar/')"
```

Avoid eval.

When a var might have spaces, double quote it.

When input is generally untrusted, protect against input that starts
with dash.

```
rm -v -- "$var"
rm -v ./* # a file could be named -rf
```

Avoid set -u. empty vars are useful, check for them with [[ $var ]]. set
-u just leads to explicitly setting variables to empty and other hackery
to avoid unset errors, making it complication that is not worth
it. shellcheck will find completely unset variables.

Prefer $var except when ${var} is necessary or adds clarity, like "comb${x}ining strings"

Never use [. [[ is a special syntax which doesn't do variable expansion
and avoids edge cases.

```
[[ $filename == *.png ]] # righthand globs over lefthand. filename has png extension
# Quote only the righthand side when globbing is not desired.
[[ $filename == "*.png" ]] # filename is literal string "*.png"
```
Use double equal for string equality [[ $x == 2 ]]



Use gnu getopts whenever there is more than one option, or an option and
an argument. Reference info at:
/usr/share/doc/util-linux/examples/getopt-parse.bash
```
usage() {
    cat <<EOF
Usage: ${0##*/} [OPTIONS] ARG1
One line description


-h|--help  Print help and exit.

Note: Uses GNU getopt options parsing style
EOF
    exit $1
}

##### begin command line parsing ########

bool_opt=false # default
long_opt=foo # default
temp=$(getopt -l help,long-opt: hso: "$@") || usage 1
eval set -- "$temp"
while true; do
    case $1 in
        -s) bool_opt=true; shift ;;
        -o|--long-opt) long_opt="$2"; shift 2 ;;
        -h|--help) usage ;;
        --) shift; break ;;
        *) echo "$0: Internal error! unexpected args: $*" ; exit 1 ;;
    esac
done
read arg1 arg2 <<<"$@"

##### end command line parsing ########
```
